#!/usr/bin/python
## Wrapper script for computing the smallest switching lattice.
import subprocess
import sys
import option
from tempfile import TemporaryFile
from espresso import dual_prods
from bounds import get_bounds
import paths
import time
import datetime
from math import ceil, floor 

start_time1=time.time()

## Option parsing
(OPT, REST) = (0, 1)
class Opt:
  mode = OPT
  area = None
  decode = False

spec = {
  "-b" : ("area", int),
  "-rest" : option.setopt("mode", REST),
  "-decode" : option.setopt("decode", True)
}

def build_lattice(infile, w, h, resfile):
  infile.seek(0)
  #start_time_sat_and_prep=time.time()
  ret = subprocess.call("./lat_to_sat.py {0} {1} | {2}".format(w, h, paths.sat),
      stdin = infile, stdout = resfile, shell = True)
  #end_time_sat_and_prep=time.time()
  #print >> sys.stderr, end_time_sat_and_prep-start_time_sat_and_prep," :SAT and prep time"
  return ret == 10

def build_lattice_rest(infile, w, h, resfile):
  infile.seek(0)
  ret = subprocess.call("./lat_to_dcsat.py {0} {1} | {2}".format(w, h, paths.sat),
      stdin = infile, stdout = resfile, shell = True)
  return ret == 10

## Find the input function
args = option.parse_opts(sys.argv[1:], spec, Opt)


print >> sys.stderr, "\n\n", sys.argv[-1]
print "\n\n", sys.argv[-1]
if len(args) > 0:
  infile = open(sys.argv[-1])
else:
  infile = sys.stdin

## Compute the upper bound on circuit size.
prods = [ elt[0] for elt in map(lambda line: line.strip().split(), infile.readlines()) ]

duals = dual_prods(prods)

assert len(prods) > 0
w_ub = 2*len(prods) - 1
h_ub = max([len(prod.replace("-","")) for prod in prods])

w_lb=h_ub
h_lb = max([len(dual.replace("-","")) for dual in duals])

#print >> sys.stderr, w_lb, h_lb


num_in = max([len(prod) for prod in prods])

ave_area = sum([len(prod.replace("-","")) for prod in prods])
w_ave = int(floor(ave_area / len(prods)))
h_ave = len(prods)

ave_area_d = sum([len(dual.replace("-","")) for dual in duals])
h_ave_d = int(floor(ave_area_d / len(duals)))
w_ave_d = len(duals)
#print >> sys.stderr,prods, ave_area,w_ave, h_ave

#w_ave=int(ceil(( w_lb + w_ave_d) /2))
#h_ave_d=int(ceil(( h_lb + h_ave)/2))

if Opt.area is None:
  Opt.area = min(w_ub*h_ub, len(prods)*len(duals)) 
  
if w_ub*h_ub<len(prods)*len(duals):
	wr=w_ub
	hr=h_ub
else:
	wr=len(prods)
	hr=len(duals)

failed = []
#configs = [(2, 2)]
configs = [(w, a/w) for a in xrange(Opt.area, 1, -1) for w in xrange(1, a) if a%w == 0]

def cmp_config((wx, hx), (wy, hy)):
  da = (wy*hy) - (wx*hx)
  if da != 0:
    return -da
  
  dasp = (wx+hx) - (wy+hy)
  if dasp != 0:
    return dasp

  return wy - wx
  
def cmp_config_2((wx, hx), (wy, hy)):
  da = (wy*hy) - (wx*hx)
  if da != 0:
    return da
  
  dasp = (wx+hx) - (wy+hy)
  if dasp != 0:
    return dasp

  return wy - wx

configs = sorted(configs, cmp_config)

best = None
res_file = TemporaryFile()

"""
for (w, h) in configs:
  if w*h > Opt.area or any([(cw >= w and ch >= h) for (cw, ch) in failed]):
    continue

  print >> sys.stderr, "Trying:", (w, h)
  if build_lattice(infile, w, h, res_file):
    print >> sys.stderr, "  Okay"
    best = (w, h, res_file)
    res_file = TemporaryFile()
    Opt.area = w*h - 1
  else:
    print >> sys.stderr, "  Failed"
    failed.append( (w, h) )
    res_file.seek(0)
"""
## Alternative version - dichotomic search.
(_, _, _, alow) = get_bounds(prods, duals)
ahigh = Opt.area-1

"""
i=-1
while i <= -1:
  i=i+1
  configs=[(h_ub,len(prods))]
  amid = (alow + ahigh)/2
"""

"""
while alow <= ahigh:
  amid = (alow + ahigh)/2
  
  configs = [ (w, amid/w) for w in xrange(1, amid+1) if (w+1)*(amid/w) > amid and w*(amid/w) <= ahigh ]
  configs = sorted(configs, cmp_config)
"""
w_ok=w_ave_d
h_ok=h_ave


with open('opt_glucose_fix_heu_bench_pap1.txt', 'a') as f:
	ts = time.time()
	st = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S')
	f.write("\n%s " % sys.argv[-1])
	f.write("%s " % st)

print >> sys.stderr, st
#with open('opt_glucose_fix_str1_once_ista6_bench.txt', 'a') as f:
#	f.write("%s " % sys.argv[-1])
i=-1;
gr=0
while i <= -1 and alow <= ahigh:
  gr=1
  amid = (alow + ahigh)/2
  configs1 = [ (w, amid/w) for w in xrange(1, amid+1) if (w+1)*(amid/w) > amid and w*(amid/w) <= ahigh ]
  configs = sorted(configs1, cmp_config)
  #print configs1
  #print configs
  if i==-1:
	configs=[(w_lb,h_lb),(w_lb,h_ave),(w_ave,h_lb),(w_ave,h_ave),(w_ave_d,h_ave_d),(w_ave,h_ave_d), (w_lb,h_ave_d),(w_ave_d,h_lb)]
	#print configs
	configs1 = sorted(configs, cmp_config)
	#print configs1
  found = False
  j=0
  for (w, h) in configs1:
    j=j+1;
    #with open('opt_glucose_fix_str1_once_ista6_bench.txt', 'a') as f:
	#	f.write("j%d " %j)
		
    if (w>=w_ave_d and h>=h_ave) or (w>=w_ok and h>=h_ok):
	  #with open('opt_glucose_fix_str1_once_ista6_bench.txt', 'a') as f:
	  #  f.write("1 ")
	  continue

    if any([(cw >= w and ch >= h) for (cw, ch) in failed]):
	  #with open('opt_glucose_fix_str1_once_ista6_bench.txt', 'a') as f:
	  #  f.write("0 ")
	  continue
      
    if (w_lb>2 and w<=2) or (w_lb==2 and w != 2) or (w_lb==1 and w != 1):
      #failed.append( (w, h) )
      #res_file.seek(0)
      #result = False
      #with open('opt_glucose_fix_str1_once_ista6_bench.txt', 'a') as f:
	  #  f.write("0 ")
      continue

    if (h_lb>2 and h<=2) or (h_lb==2 and h != 2) or (h_lb==1 and h != 1):
      #failed.append( (w, h) )
      #res_file.seek(0)
      #result = False
	  #with open('opt_glucose_fix_str1_once_ista6_bench.txt', 'a') as f:
	  #  f.write("0 ")
	  continue


    print  "Trying:", (w, h)
    if Opt.mode == OPT:
      result = build_lattice(infile, w, h, res_file)
    else:
      result = build_lattice_rest(infile, w, h, res_file)
    if result:
      print "  Okay"
      #with open('opt_glucose_fix_str1_once_ista6_bench.txt', 'a') as f:
	  #  f.write("1 ")
      best = (w, h, res_file)
      res_file = TemporaryFile()
      ahigh = w*h - 1
      w_ok=w
      h_ok=h
      found = True
      break
    else:
      print  "  Failed"
      #with open('opt_glucose_fix_str1_once_ista6_bench.txt', 'a') as f:
	  #  f.write("0 ")
      failed.append( (w, h) )
      res_file.seek(0)
  i=i+1
  """
  found = False
  j=0
  for (w, h) in configs:
    j=j+1;
    #with open('opt_glucose_fix_str1_once_ista6_bench.txt', 'a') as f:
	#	f.write("j%d " %j)
		
    if (w>=w_ave_d and h>=h_ave) or (w>=w_ok and h>=h_ok):
	  with open('opt_glucose_fix_str1_once_ista6_bench.txt', 'a') as f:
	    f.write("1 ")
	  continue

    if any([(cw >= w and ch >= h) for (cw, ch) in failed]):
	  with open('opt_glucose_fix_str1_once_ista6_bench.txt', 'a') as f:
	    f.write("0 ")
	  continue
      
    if (w_lb>2 and w<=2) or (w_lb==2 and w != 2) or (w_lb==1 and w != 1):
      #failed.append( (w, h) )
      #res_file.seek(0)
      #result = False
      with open('opt_glucose_fix_str1_once_ista6_bench.txt', 'a') as f:
	    f.write("0 ")
      continue

    if (h_lb>2 and h<=2) or (h_lb==2 and h != 2) or (h_lb==1 and h != 1):
      #failed.append( (w, h) )
      #res_file.seek(0)
      #result = False
	  with open('opt_glucose_fix_str1_once_ista6_bench.txt', 'a') as f:
	    f.write("0 ")
	  continue


    print  "Trying:", (w, h)
    if Opt.mode == OPT:
      result = build_lattice(infile, w, h, res_file)
    else:
      result = build_lattice_rest(infile, w, h, res_file)
    if result:
      print "  Okay"
      with open('opt_glucose_fix_str1_once_ista6_bench.txt', 'a') as f:
	    f.write("1 ")
      best = (w, h, res_file)
      res_file = TemporaryFile()
      ahigh = w*h - 1
      #w_ok=w
      #h_ok=h
      found = True
      #break
    else:
      print  "  Failed"
      with open('opt_glucose_fix_str1_once_ista6_bench.txt', 'a') as f:
	    f.write("0 ")
      failed.append( (w, h) )
      res_file.seek(0)
  """
  #with open('opt_glucose_fix_str1_once_ista6_bench.txt', 'a') as f:
	#f.write("\n")
  if not found:
    alow = amid+1
#"""

if best is not None:
  (w, h, best_file) = best
  print "Best:", (w, h)

  for line in best_file.readlines():
    print line 
  if Opt.decode:
    best_file.seek(0)
    subprocess.call(["./decode.py", str(num_in), str(w), str(h)], stdin = best_file)

infile.close()

end_time1=time.time()
tot_time=end_time1-start_time1
print >> sys.stderr, tot_time," :total elapsed time","\n"

tot_time=str(tot_time)
"""
with open('fixed_pseudo_UB_bench.txt', 'a') as f:
	f.write("%s " % sys.argv[-1])
	if best is not None:
		(w, h, best_file) = best
		f.write("%s " % tot_time)
		f.write("Real_UB %d " % (Opt.area+1))
		f.write("Pseudo_UB %d " % (w*h))
		f.write("Okay ")
		f.write("%d" % w)
		f.write("x")
		f.write("%d\n" % h)
	else:
		f.write("%s\n" % tot_time)
		f.write("Real_UB %d " % (Opt.area+1))
		f.write("Pseudo_UB %d " % (len(prods)*h_ub))
		f.write("Failed ")
		f.write("%d" % h_ub)
		f.write("x")
		f.write("%d\n" % len(prods))"""

"""
with open('opt_glucose_fix_str1_once_ista9_bench.txt', 'a') as f:
	f.write("%s " % sys.argv[-1])
	if best is not None:
		(w, h, best_file) = best
		f.write("%s " % tot_time)
		f.write("%d" % w)
		f.write("x")
		f.write("%d " % h)
		f.write("w_lb %d " % w_lb)
		f.write("h_lb %d " % h_lb)
		f.write("w_ave %d " % w_ave)
		f.write("h_ave %d " % h_ave)
		f.write("w_ave_d %d " % w_ave_d)
		f.write("h_ave_d %d " % h_ave_d)
		
		if w<=w_lb and h<=h_lb:
			f.write("1 ")
		else:
			f.write("0 ")
			
		if w<=w_lb and h<=h_ave:
			f.write("1 ")
		else:
			f.write("0 ")
			
		if w<=w_ave and h<=h_lb:
			f.write("1 ")
		else:
			f.write("0 ")
			
		if w<=w_ave and h<=h_ave:
			f.write("1 ")
		else:
			f.write("0 ")
			
		if w<=w_ave_d and h<=h_ave_d:
			f.write("1 ")
		else:
			f.write("0 ")
			
		if w<=w_ave and h<=h_ave_d:
			f.write("1 ")
		else:
			f.write("0 ")
			
		if w<=w_lb and h<=h_ave_d:
			f.write("1 ")
		else:
			f.write("0 ")
			
		if w<=w_ave_d and h<=h_lb:
			f.write("1 ")
		else:
			f.write("0 ")
		
		
		f.write("optimum_check ")
	
	
		if w*h<=w_lb*h_lb:
			f.write("1 ")
		else:
			f.write("0 ")
			
		if w*h<=w_lb*h_ave:
			f.write("1 ")
		else:
			f.write("0 ")
			
		if w*h<=w_ave*h_lb:
			f.write("1 ")
		else:
			f.write("0 ")
			
		if w*h<=w_ave*h_ave:
			f.write("1 ")
		else:
			f.write("0 ")
			
		if w*h<=w_ave_d*h_ave_d:
			f.write("1 ")
		else:
			f.write("0 ")
			
		if w*h<=w_ave*h_ave_d:
			f.write("1 ")
		else:
			f.write("0 ")
			
		if w*h<=w_lb*h_ave_d:
			f.write("1 ")
		else:
			f.write("0 ")
			
		if w*h<=w_ave_d*h_lb:
			f.write("1\n")
		else:
			f.write("0\n")
	else:
		f.write("%s\n" % tot_time)
"""

with open('opt_glucose_fix_heu_bench_pap1.txt', 'a') as f:
	#f.write("%s " % sys.argv[-1])
	if best is not None:
		(w, h, best_file) = best
		f.write("%s " % tot_time)
		f.write("%d" % w)
		f.write("x")
		f.write("%d " % h)
		f.write("%d " % w)
		f.write("%d" % h)
	elif gr==0:
		f.write("%s " % tot_time)
		f.write("%d" % hr)
		f.write("x")
		f.write("%d " % wr)
		f.write("%d " % hr)
		f.write("%d" % wr)
	else:
		f.write("%s" % tot_time)
