#!/usr/bin/python

## Code for converting a circuit into a SAT instance
## which produces an n x m lattice, by unfolding the
## quantified variables.
import sys
import qbf
from constraints import sorter, cardnet, atmost_one, reify #, lex_leq
from formula import ExpTable, BoolExp
from itertools import chain
from espresso import dual_prods
import re
import time

#start_time_prep2=time.time()
target_idx = 0

ncols = int(sys.argv[-1])
nrows = int(sys.argv[-2])

problem = qbf.QBF()
if nrows>1 and ncols>1:
	pos_name="../lib_sab/{}x{}_pos.txt".format(nrows, ncols)
	pos_profile = open(pos_name, "r")
	sab_pos_clauses = pos_profile.readlines()

	neg_name="../lib_sab/{}x{}_neg.txt".format(nrows, ncols)
	neg_profile = open(neg_name, "r")
	sab_neg_clauses = neg_profile.readlines()

symbreak = False
#if "-sym" in sys.argv[1:-2]:
#  symbreak = True
#symbreak = True

## Assume input is in pla format.
invars = set()

prods = []
for line in sys.stdin.readlines():
  [sin, sout] = line.strip().split()
  if sout[target_idx] == '1':
    prods.append(sin)
    for ti in xrange(len(sin)):
      vi = ti+1
      if sin[ti] == '0' or sin[ti] == '1':
        invars |= {vi}

## Convert the set of input variables
## to a list.
invars = sorted(list(invars))

## Trim the products to the care-set
prods = map(lambda prod: "".join([prod[vi-1] for vi in invars]), prods)
duals = dual_prods(prods)


## Set up the expression table
table = ExpTable()
def toExp(v):
  return table.leaf(v)

## Construct the quantified variables
## Ordering [F, T, x0, ~x0, x1, ~x1, ...]
latvars = { (i, j) : problem.existFD(2 + 2*len(invars))
              for i in xrange(nrows) for j in xrange(ncols) }
## qvars = { v : problem.univ() for v in invars }

## Symmetry breaking
##if symbreak:
##  left_vars = [ latvars[(i, j)] for i in xrange(nrows) for j in xrange(ncols/2) ]
##  right_vars = [ latvars[(i, ncols-j-1)] for i in xrange(nrows) for j in xrange(ncols/2) ]
##
##  lex_leq(problem, left_vars, right_vars)
##
##  top_vars = [ latvars[(i, j)] for j in xrange(ncols) for i in xrange(nrows/2) ]
##  bot_vars = [ latvars[(nrows-i-1, j)] for j in xrange(ncols) for i in xrange(nrows/2) ]
##
##  lex_leq(problem, top_vars, bot_vars)

#for (i, j) in latvars:
#  print >> sys.stderr, (i, j), ":", latvars[(i, j)]

## Collect the path variables adjacent to a square.
def pos_adjvars(vs, i, j):
  if i == 0:
    return [ vs[(i+1, j)] ]
  if i == nrows-1:
    return [ vs[(i-1, j)] ]
  return [ vs[elt] for elt in [(i-1, j), (i+1, j), (i, j-1), (i, j+1)] if elt in vs ]

def neg_adjvars(vs, i, j):
  if j == 0:
    return [ vs[(ip, j+1)] for ip in [i-1, i, i+1] if (ip, j+1) in vs ]
  if j == ncols-1:
    return [ vs[(ip, j-1)] for ip in [i-1, i, i+1] if (ip, j-1) in vs ]
  return [ vs[elt] for elt in 
            [(i-1, j-1), (i-1, j), (i-1, j+1), (i, j-1), (i, j+1), (i+1, j-1), (i+1, j), (i+1,j+1)]
            if elt in vs ]

## Build the positive instances for a given assignment
def build_pos(prod):
  #print >> sys.stderr, "prod umuz", prod, nrows, ncols, "\n"
  nvars = { (i, j): problem.exist() for i in xrange(nrows) for j in xrange(ncols) }
  #print >> sys.stderr, "neyse bu nvar:", nvars, "\n"
  for (i, j) in nvars.keys():
    nv = nvars[(i, j)]
    #print >> sys.stderr, "ne ki nv:", nv
    ## ~switch_{ij} => ~nv
    lvar = latvars[(i, j)].vs
    #print >> sys.stderr, "ne bu ki lvar:", lvar, "\n"
    assert len(lvar) == 2*len(prod) + 2

#    """
    problem.addClause([~lvar[0], ~nv])
    for vi in xrange(len(prod)):
      if prod[vi] == '0':
        #print >> sys.stderr, "prod vi 0",vi,"-", (~lvar[2*vi + 2]), ~nv, "\n"
        problem.addClause([~lvar[2*vi + 2], ~nv])
      elif prod[vi] == '1':
        #print >> sys.stderr, "prod vi 1",vi,"-", (~lvar[2*vi + 3]), ~nv, "\n"
        problem.addClause([~lvar[2*vi + 3], ~nv])
      else:
        # Don't-cares can't be used
        problem.addClause([~lvar[2*vi + 2], ~nv])
        problem.addClause([~lvar[2*vi + 3], ~nv])

    """
    sv = problem.exist()
    problem.addClause([~lvar[0], ~sv])
    problem.addClause([~lvar[1], sv])
    for vi in xrange(len(prod)):
      if prod[vi] == '0':
        problem.addClause([~lvar[2*vi + 2], ~sv])
        problem.addClause([~lvar[2*vi + 3], sv])
      elif prod[vi] == '1':
        problem.addClause([~lvar[2*vi + 3], ~sv])
        problem.addClause([~lvar[2*vi + 2], sv])
      else:
        # Don't-cares can't be used
        problem.addClause([~lvar[2*vi + 2], ~sv])
        problem.addClause([~lvar[2*vi + 3], ~sv])
#    problem.addClause([~sv, lvar[1]]
#        + [ lvar[2*vi + 2] for vi in xrange(len(prod)) if prod[vi] == '1' ]
#        + [ lvar[2*vi + 3] for vi in xrange(len(prod)) if prod[vi] == '0' ])

    problem.addClause([sv, ~nv])
#  """
  nv00 = nvars[(0,0)]
  nv00 = str(nv00)
  nv00 = int(nv00)-1
  if nrows == 1:
    problem.addClause(nvars.values())
    #print >> sys.stderr, "ekle nrow 1", nvars.values(), "\n"
  elif ncols == 1:
    for val in nvars.values():
      problem.addClause([val])
      #print >> sys.stderr, "ekle ncol 1", [val], "\n"
  else:
    for i in range(0,len(sab_pos_clauses)):
	  sab_pos_clause = sab_pos_clauses[i]
	  sab_pos_clause=map(int, re.findall(r'\d+', sab_pos_clause))
	  asg_pos_clause=map(lambda x:x+nv00, sab_pos_clause)
	  #print asg_pos_clause
	  problem.addClause(asg_pos_clause)
 
"""bura yok
    for (i, j) in nvars.keys():
      #print >> sys.stderr, "nvar i j", (i,j), "\n"
      nv = nvars[(i, j)]
      #print >> sys.stderr, "nvli", nv, "\n"
      adjs = pos_adjvars(nvars, i, j)
      #print >> sys.stderr, "adjs", adjs, "\n"
      assert len(adjs) in [1, 3, 4]
      
      if len(adjs) == 1:
        ## Terminal edge.
        problem.addClause([adjs[0],~nv])
      else:
        assert len(adjs) >= 3
        [s1, s2, s3] = cardnet(problem, adjs, 3)
        #print >> sys.stderr, "s ler", s1, s2, s3, "\n"
        ## nv -> sum(adjs) = 2
        problem.addClause([s2, ~nv])
        problem.addClause([~s3, ~nv])

    ## Exactly one source
    atmost_one(problem, [ nvars[(i, j)] for (i, j) in nvars.keys() if i == 0 ], problem.ttt())
    ## Exactly one destination
    atmost_one(problem, [ nvars[(i, j)] for (i, j) in nvars.keys() if i == nrows-1 ], problem.ttt())
"""

## And the negative paths
def build_neg(prod):
  #print >> sys.stderr, "prod umuz", prod, nrows, ncols, "\n"
  nvars = { (i, j): problem.exist() for i in xrange(nrows) for j in xrange(ncols) }
  #print >> sys.stderr, "neyse bu nvar:", nvars, "\n"
  for (i, j) in nvars.keys():
    nv = nvars[(i, j)]
    #print >> sys.stderr, "ne ki nv:", nv
    ## switch_{ij} => ~nv
    lvar = latvars[(i, j)].vs
    #print >> sys.stderr, "ne bu ki lvar:", lvar, "\n"
    assert len(lvar) == 2*len(prod) + 2

#    """
    problem.addClause([~lvar[1], ~nv]) ## [[ lij = T ]] => ~neg[i,j]
    for vi in xrange(len(prod)):
      if prod[vi] == '0':
        #print >> sys.stderr, "prod vi 0",vi,"-", (~lvar[2*vi + 3]), ~nv, "\n"
        problem.addClause([~lvar[2*vi + 3], ~nv])
      elif prod[vi] == '1':
        #print >> sys.stderr, "prod vi 1",vi,"-", (~lvar[2*vi + 3]), ~nv, "\n"
        problem.addClause([~lvar[2*vi + 2], ~nv])
      else:
        # Don't care for this product
        problem.addClause([~lvar[2*vi + 3], ~nv])
        problem.addClause([~lvar[2*vi + 2], ~nv])
    """
    sv = problem.exist()
    problem.addClause([~lvar[0], sv])
    problem.addClause([~lvar[1], ~sv])
    for vi in xrange(len(prod)):
      if prod[vi] == '0':
        problem.addClause([~lvar[2*vi + 3], ~sv])
        problem.addClause([~lvar[2*vi + 2], sv])
      elif prod[vi] == '1':
        problem.addClause([~lvar[2*vi + 2], ~sv])
        problem.addClause([~lvar[2*vi + 3], sv])
      else:
        # Don't-cares can't be used
        problem.addClause([~lvar[2*vi + 3], ~sv])
        problem.addClause([~lvar[2*vi + 2], ~sv])
#    problem.addClause([~sv, lvar[0]]
#        + [ lvar[2*vi + 3] for vi in xrange(len(prod)) if prod[vi] == '1' ]
#        + [ lvar[2*vi + 2] for vi in xrange(len(prod)) if prod[vi] == '0' ])

    problem.addClause([sv, ~nv])
#    """

  #print >> sys.stderr, "nvar keys", nvars.keys()
  #print >> sys.stderr, "nvar values", nvars.values(), "\n"
  nv00 = nvars[(0,0)]
  nv00 = str(nv00)
  nv00 = int(nv00)-1
  if ncols == 1:
    problem.addClause(nvars.values())
    #print >> sys.stderr, "ekle ncol 1", nvars.values(), "\n"
  elif nrows == 1:
    for val in nvars.values():
      problem.addClause([val])
      #print >> sys.stderr, "ekle nrow 1", [val], "\n"
  else:
	for i in range(0,len(sab_neg_clauses)):
	  sab_neg_clause = sab_neg_clauses[i]
	  sab_neg_clause=map(int, re.findall(r'\d+', sab_neg_clause))
	  asg_neg_clause=map(lambda x:x+nv00, sab_neg_clause)
	  #print asg_neg_clause
	  problem.addClause(asg_neg_clause)
	  
"""
    for (i, j) in nvars.keys():
      ## Connect up the path variables.
      #print >> sys.stderr, "nvar i j", (i,j), "\n"
      nv = nvars[(i, j)]
      #print >> sys.stderr, "nvli", nv, "\n"
      adjs = neg_adjvars(nvars, i, j)
      #print >> sys.stderr, "adjs", adjs, "\n"
      assert (len(adjs) in [2, 3, 5, 8])
      
      if len(adjs) <= 3:
        ## Terminal edge.
        assert j in [0, ncols-1]
        adjv = problem.exist()
        atmost_one(problem, adjs, adjv)
        problem.addClause([adjv, ~nv])
      else:
        [s1, s2, s3] = cardnet(problem, adjs, 3)
        problem.addClause([s2, ~nv])
        problem.addClause([~s3, ~nv])

    ## Exactly one source
    atmost_one(problem, [ nvars[(i, j)] for (i, j) in nvars.keys() if j == 0 ], problem.ttt())
    ## Exactly one destination
    atmost_one(problem, [ nvars[(i, j)] for (i, j) in nvars.keys() if j == ncols-1 ], problem.ttt())
"""

for prod in prods:
  build_pos(prod)

for prod in duals:
  build_neg(prod)

problem.to_dimacs()
