#!/usr/bin/python

## Code for converting a circuit into a SAT instance
## which produces an n x m lattice, by unfolding the
## quantified variables.
import sys
import qbf
from constraints import sorter, cardnet, atmost_one, reify
from formula import ExpTable, BoolExp
from itertools import chain
import re
import time

#start_time_prep2=time.time()

target_idx = 0

ncols = int(sys.argv[-1])
nrows = int(sys.argv[-2])
#print >> sys.stderr, "2_bozuk"
problem = qbf.QBF()
if nrows>1 and ncols>1:
	pos_name="../lib_sab/{}x{}_pos.txt".format(nrows, ncols)
	pos_profile = open(pos_name, "r")
	sab_pos_clauses = pos_profile.readlines()

	neg_name="../lib_sab/{}x{}_neg.txt".format(nrows, ncols)
	neg_profile = open(neg_name, "r")
	sab_neg_clauses = neg_profile.readlines()

## Assume input is in pla format.
invars = set()
terms = []
for line in sys.stdin.readlines():
  [sin, sout] = line.strip().split()
  #print >> sys.stderr, "line giris felan", sin, sout
  ## The given minterm has f(X) = 1
  if sout[target_idx] == '1':
    term = []
    for ti in xrange(len(sin)):
      vi = ti+1
      if sin[ti] == '1':
        term.append(vi)
        invars |= { vi }
      elif sin[ti] == '0':
        term.append(-vi)
        invars |= { vi }

    terms.append(term)
    
# print terms

## Convert the set of input variables
## to a list.

invars = list(invars)

## Set up the expression table
table = ExpTable()

def toExp(v):
  return table.leaf(v)

## Construct the quantified variables
## Ordering [F, T, x0, ~x0, x1, ~x1, ...]
latvars = { (i, j) : problem.existFD(2 + 2*len(invars))
              for i in xrange(nrows) for j in xrange(ncols) }
              
#print >> sys.stderr, "latice var:", latvars

## qvars = { v : problem.univ() for v in invars }

#for (i, j) in latvars:
#  print >> sys.stderr, (i, j), ":", latvars[(i, j)]

## Collect the path variables adjacent to a square.
def pos_adjvars(vs, i, j):
  if i == 0:
    return [ vs[(i+1, j)] ]
  if i == nrows-1:
    return [ vs[(i-1, j)] ]
  return [ vs[elt] for elt in [(i-1, j), (i+1, j), (i, j-1), (i, j+1)] if elt in vs ]

def neg_adjvars(vs, i, j):
  if j == 0:
    return [ vs[(ip, j+1)] for ip in [i-1, i, i+1] if (ip, j+1) in vs ]
  if j == ncols-1:
    return [ vs[(ip, j-1)] for ip in [i-1, i, i+1] if (ip, j-1) in vs ]
  return [ vs[elt] for elt in 
            [(i-1, j-1), (i-1, j), (i-1, j+1), (i, j-1), (i, j+1), (i+1, j-1), (i+1, j), (i+1,j+1)]
            if elt in vs ]

##pos_nvars = { (i, j): problem.exist() for i in xrange(nrows) for j in xrange(ncols) }
##neg_nvars = { (i, j): problem.exist() for i in xrange(nrows) for j in xrange(ncols) }

## Build the positive instances for a given assignment
def build_pos(asg):
  #print >> sys.stderr, "asg miz", asg, nrows, ncols, "\n"
  nvars = { (i, j): problem.exist() for i in xrange(nrows) for j in xrange(ncols) }
  #print >> sys.stderr, "neyse bu nvar:", nvars, "\n"
  for (i, j) in nvars.keys():
    nv = nvars[(i, j)]
    #print >> sys.stderr, "ne ki nv:", nv
    ## ~switch_{ij} => ~nv
    lvar = latvars[(i, j)].vs
    #print >> sys.stderr, "ne bu ki lvar:", lvar, "\n"
    assert len(lvar) == 2*len(asg) + 2

    #problem.addClause([~lvar[1], ~nv])
    problem.addClause([~lvar[0], ~nv])
    #print >> sys.stderr, "eklenen", (~lvar[0]), ~nv, "\n"
    
    for vi in xrange(len(asg)):
      if asg[vi]:
        #print >> sys.stderr, "asg vi 1",vi,"-", (~lvar[2*vi + 3]), ~nv, "\n"
        problem.addClause([~lvar[2*vi + 3], ~nv])
      else:
        #print >> sys.stderr,"asg vi 0",vi,"-", (~lvar[2*vi + 2]), ~nv, "\n"
        problem.addClause([~lvar[2*vi + 2], ~nv])
  
  #print >> sys.stderr, "nvar keys", nvars.keys()
  #print >> sys.stderr, "nvar values", nvars.values(), "\n"
  
  nv00 = nvars[(0,0)]
  nv00 = str(nv00)
  nv00 = int(nv00)-1
  if nrows == 1:
    problem.addClause(nvars.values())
    #print >> sys.stderr, "ekle nrow 1", nvars.values(), "\n"
  elif ncols == 1:
    for val in nvars.values():
      problem.addClause([val])
      #print >> sys.stderr, "ekle ncol 1", [val], "\n"
  else:  
	for i in range(0,len(sab_pos_clauses)):
	  sab_pos_clause = sab_pos_clauses[i]
	  sab_pos_clause=map(int, re.findall(r'\d+', sab_pos_clause))
	  asg_pos_clause=map(lambda x:x+nv00, sab_pos_clause)
	  #print asg_pos_clause
	  problem.addClause(asg_pos_clause)
  


"""
#bura yok  
  if nrows == 1:
    problem.addClause(nvars.values())
    print >> sys.stderr, "ekle nrow 1", nvars.values(), "\n"
  elif ncols == 1:
    for val in nvars.values():
      problem.addClause([val])
      print >> sys.stderr, "ekle ncol 1", [val], "\n"
  else:
    for (i, j) in nvars.keys():
      print >> sys.stderr, "nvar i j", (i,j), "\n"
      nv = nvars[(i, j)]
      print >> sys.stderr, "nvli", nv, "\n"
      adjs = pos_adjvars(nvars, i, j)
      print >> sys.stderr, "adjs", adjs, "\n"
      assert len(adjs) in [1, 3, 4]
      
      if len(adjs) == 1:
        ## Terminal edge.
        problem.addClause([adjs[0],~nv])
      else:
        assert len(adjs) >= 3
        [s1, s2, s3] = cardnet(problem, adjs, 3)
        print >> sys.stderr, "s ler", s1, s2, s3, "\n"
        ## nv -> sum(adjs) = 2
        problem.addClause([s2, ~nv])
        problem.addClause([~s3, ~nv])
    
    print >> sys.stderr,"hopbala pos"
    #print >> sys.stderr,"hopbala","\n", problem.clauses,"\n", problem.nvars, "\n"
    ## Exactly one source
    atmost_one(problem, [ nvars[(i, j)] for (i, j) in nvars.keys() if i == 0 ], problem.ttt())
    ## Exactly one destination
    atmost_one(problem, [ nvars[(i, j)] for (i, j) in nvars.keys() if i == nrows-1 ], problem.ttt())
"""
    
## And the negative paths
def build_neg(asg):
  #print >> sys.stderr, "\n\nasg miz", asg, nrows, ncols, "\n"
  nvars = { (i, j): problem.exist() for i in xrange(nrows) for j in xrange(ncols) }
  #print >> sys.stderr, "neyse bu nvar:", nvars, "\n"
  for (i, j) in nvars.keys():
    nv = nvars[(i, j)]
    #print >> sys.stderr, "ne ki nv:", nv
    ## switch_{ij} => ~nv
    lvar = latvars[(i, j)].vs
    #print >> sys.stderr, "ne bu ki lvar:", lvar, "\n"
    assert len(lvar) == 2*len(asg) + 2

    problem.addClause([~lvar[1], ~nv]) ## [[ lij = T ]] => ~neg[i,j]
    #problem.addClause([~lvar[0], ~nv])
    #print >> sys.stderr, "eklenen", (~lvar[0]), ~nv, "\n"
    
    for vi in xrange(len(asg)):
      if asg[vi]:
        problem.addClause([~lvar[2*vi + 2], ~nv])
        #print >> sys.stderr, "asg vi 1",vi,"-", (~lvar[2*vi + 2]), ~nv, "\n"
      else:
        problem.addClause([~lvar[2*vi + 3], ~nv])
        #print >> sys.stderr, "asg vi 0",vi,"-", (~lvar[2*vi + 3]), ~nv, "\n"
    
  nv00 = nvars[(0,0)]
  nv00 = str(nv00)
  nv00 = int(nv00)-1
  if ncols == 1:
    problem.addClause(nvars.values())
    #print >> sys.stderr, "ekle ncol 1", nvars.values(), "\n"
  elif nrows == 1:
    for val in nvars.values():
      problem.addClause([val])
      #print >> sys.stderr, "ekle nrow 1", [val], "\n"
  else:  
	for i in range(0,len(sab_neg_clauses)):
	  sab_neg_clause = sab_neg_clauses[i]
	  sab_neg_clause=map(int, re.findall(r'\d+', sab_neg_clause))
	  asg_neg_clause=map(lambda x:x+nv00, sab_neg_clause)
	  #print asg_neg_clause
	  problem.addClause(asg_neg_clause)
  
  

"""
#bura yok
  if ncols == 1:
    problem.addClause(nvars.values())
    print >> sys.stderr, "ekle ncol 1", nvars.values(), "\n"
  elif nrows == 1:
    for val in nvars.values():
      problem.addClause([val])
      print >> sys.stderr, "ekle nrow 1", [val], "\n"
  else:
    for (i, j) in nvars.keys():
      ## Connect up the path variables.
      print >> sys.stderr, "nvar ij", (i,j), "\n"
      nv = nvars[(i, j)]
      print >> sys.stderr, "nvli", nv, "\n"
      adjs = neg_adjvars(nvars, i, j)
      print >> sys.stderr, "adjs", adjs, nvars, i ,j, "\n"
      assert (len(adjs) in [2, 3, 5, 8])
      
      if len(adjs) <= 3:
        ## Terminal edge.
        assert j in [0, ncols-1]
        adjv = problem.exist()
        atmost_one(problem, adjs, adjv)
        problem.addClause([adjv, ~nv])
      else:
        [s1, s2, s3] = cardnet(problem, adjs, 3)
        problem.addClause([s2, ~nv])
        problem.addClause([~s3, ~nv])
    
    print >> sys.stderr,"hopbala neg"
    #print >> sys.stderr, "hopbala", "\n", problem.clauses(),"\n", problem.nvars(), "\n",problem.ttt()
    ## Exactly one source
    atmost_one(problem, [ nvars[(i, j)] for (i, j) in nvars.keys() if j == 0 ], problem.ttt())
    ## Exactly one destination
    atmost_one(problem, [ nvars[(i, j)] for (i, j) in nvars.keys() if j == ncols-1 ], problem.ttt())
"""

def all_assigns(k):
  if k <= 0:
    return [ [] ]
  else:
    rest = all_assigns(k-1)
    return map(lambda asg: [False] + asg, rest) + map(lambda asg: [True] + asg, rest)

def eval_assign(asg):
  def eval_elt(elt, asg):
    if elt > 0:
#      print elt
      return asg[elt-1]
    else:
      return not asg[-elt-1]
  def eval_term(term, asg):
    return reduce(lambda x, y: x and eval_elt(y, asg), term, True)

  return reduce(lambda x, y: x or eval_term(y, asg), terms, False) 

for asg in all_assigns(len(invars)):
#  print >> sys.stderr, asg
#  print >> sys.stderr, terms
#  print >> sys.stderr, eval_assign(asg)
  if eval_assign(asg):
#    print >> sys.stderr, "T:", asg
    build_pos(asg)
  else:
#    print >> sys.stderr, "F:", asg
    build_neg(asg)
#print >> sys.stderr, "cuppala"


problem.to_dimacs()

#end_time3=time.time()
#print >>sys.stderr, end_time3-start_time_prep2, " :time prepare input to SAT for optimal"
