#!/usr/bin/python
## Wrapper script for computing the smallest switching lattice.
import subprocess
import sys
import option
from tempfile import TemporaryFile
from espresso import dual_prods
from bounds import get_bounds
import paths
import time
start_time1=time.time()

## Option parsing
(OPT, REST) = (0, 1)
class Opt:
  mode = OPT
  area = None
  decode = False

spec = {
  "-b" : ("area", int),
  "-rest" : option.setopt("mode", REST),
  "-decode" : option.setopt("decode", True)
}

def build_lattice(infile, w, h, resfile):
  infile.seek(0)
  #start_time_sat_and_prep=time.time()
  ret = subprocess.call("./lat_to_sat.py {0} {1} | {2}".format(w, h, paths.sat),
      stdin = infile, stdout = resfile, shell = True)
  #end_time_sat_and_prep=time.time()
  #print >> sys.stderr, end_time_sat_and_prep-start_time_sat_and_prep," :SAT and prep time"
  return ret == 10

def build_lattice_rest(infile, w, h, resfile):
  infile.seek(0)
  ret = subprocess.call("./lat_to_dcsat.py {0} {1} | {2}".format(w, h, paths.sat),
      stdin = infile, stdout = resfile, shell = True)
  return ret == 10

## Find the input function
args = option.parse_opts(sys.argv[1:], spec, Opt)

print >> sys.stderr, "\n\n", sys.argv[-1]
if len(args) > 0:
  infile = open(sys.argv[-1])
else:
  infile = sys.stdin

## Compute the upper bound on circuit size.
prods = [ elt[0] for elt in map(lambda line: line.strip().split(), infile.readlines()) ]
#print >> sys.stderr, "prods:",prods
duals = dual_prods(prods)
#print >> sys.stderr, "duals:",duals

assert len(prods) > 0
w_ub = 2*len(prods) - 1
h_ub = max([len(prod.replace("-","")) for prod in prods])

num_in = max([len(prod) for prod in prods])

if Opt.area is None:
  Opt.area = min(w_ub*h_ub, len(prods)*len(duals)) -1

#Opt.area=25
failed = []
#configs = [(2, 2)]
configs = [(w, a/w) for a in xrange(Opt.area, 1, -1) for w in xrange(1, a) if a%w == 0]

def cmp_config((wx, hx), (wy, hy)):
  da = (wy*hy) - (wx*hx)
  if da != 0:
    return da
  
  dasp = (wx+hx) - (wy+hy)
  if dasp != 0:
    return dasp

  return wy - wx

configs = sorted(configs, cmp_config)

best = None
res_file = TemporaryFile()

"""
for (w, h) in configs:
  if w*h > Opt.area or any([(cw >= w and ch >= h) for (cw, ch) in failed]):
    continue

  print >> sys.stderr, "Trying:", (w, h)
  if build_lattice(infile, w, h, res_file):
    print >> sys.stderr, "  Okay"
    best = (w, h, res_file)
    res_file = TemporaryFile()
    Opt.area = w*h - 1
  else:
    print >> sys.stderr, "  Failed"
    failed.append( (w, h) )
    res_file.seek(0)
"""
## Alternative version - dichotomic search.
(_, _, _, alow) = get_bounds(prods, duals)
ahigh = Opt.area

"""
i=-1
while i <= -1:
  i=i+1
  configs=[(3,3)]
  amid = (alow + ahigh)/2
"""

"""
while alow <= ahigh:
  amid = (alow + ahigh)/2
  
  configs = [ (w, amid/w) for w in xrange(1, amid+1) if (w+1)*(amid/w) > amid and w*(amid/w) <= ahigh ]
  configs = sorted(configs, cmp_config)
"""
while alow <= ahigh:
  amid = (alow + ahigh)/2
  
  configs = [ (w, amid/w) for w in xrange(1, amid+1) if (w+1)*(amid/w) > amid and w*(amid/w) <= ahigh ]
  configs = sorted(configs, cmp_config)

  found = False
  for (w, h) in configs:
    #(w,h)=configs[i] 
    if any([(cw >= w and ch >= h) for (cw, ch) in failed]):
      continue

    print >> sys.stderr, "Trying:", (w, h)
    if Opt.mode == OPT:
      result = build_lattice(infile, w, h, res_file)
    else:
      result = build_lattice_rest(infile, w, h, res_file)
    if result:
      print >> sys.stderr, "  Okay"
      best = (w, h, res_file)
      res_file = TemporaryFile()
      ahigh = w*h - 1
      found = True
      break
    else:
      print >> sys.stderr, "  Failed"
      failed.append( (w, h) )
      res_file.seek(0)

  if not found:
    alow = amid+1
#"""

if best is not None:
  (w, h, best_file) = best
  print "Best:", (w, h)

  for line in best_file.readlines():
    print line 
  if Opt.decode:
    best_file.seek(0)
    subprocess.call(["./decode.py", str(num_in), str(w), str(h)], stdin = best_file)

infile.close()

end_time1=time.time()

tot_time=end_time1-start_time1
print >> sys.stderr, tot_time," :total elapsed time","\n"

tot_time=str(tot_time)
with open('8bozuk_glucose_bench.txt', 'a') as f:
	f.write("%s " % sys.argv[-1])
	if best is not None:
		(w, h, best_file) = best
		f.write("%s " % tot_time)
		f.write("%d" % w)
		f.write("x")
		f.write("%d\n" % h)
	else:
		f.write("%s\n" % tot_time)





