## Module for constructing QBF formulae.
## Only Booleans may be Univ. quantified;
## FD variables must be Ex.

## Simple structurally-hashed Boolean-formula
## table.
from lit import Lit, FDVar
from constraints import atmost_one
from formula import ExpTable, BoolExp
import sys

class QBF:
  (EXIST, UNIV) = (0, 1)
  QF_repr = { EXIST: 'e', UNIV: 'a' }

  def __init__(self):
    self.blocks = [ (QBF.EXIST, [1]) ]
##    self.formulae = []
    self.clauses = [ [Lit(1)] ]
    self.nvars = 1

  def ttt(self):
    return Lit(1)
  def fff(self):
    return Lit(-1)

  ## Should do variable substitution.
  ## Instead, just adding x <=> y.
  def unify(self, x, y):
    self.addClause([x, ~y])
    self.addClause([y, ~x])

  def exist(self):
    self.nvars += 1
    v = self.nvars

    if len(self.blocks) == 0 or self.blocks[-1][0] == QBF.UNIV:
      self.blocks.append( (QBF.EXIST, []) )
    self.blocks[-1][1].append(v)
    
    return Lit(v)

  def univ(self):
    self.nvars += 1
    v = self.nvars
    if len(self.blocks) == 0 or self.blocks[-1][0] == QBF.EXIST:
      self.blocks.append( (QBF.UNIV, []) )
    self.blocks[-1][1].append(v)

    return Lit(v)

  ## x <- [low, high] (Closed interval)
  def existFD(self, dom):
    assert dom >= 1

    if dom == 1:
      return [ self.ttt() ]
    elif dom == 2:
      v = self.exist()
      return [~v, v]
    else:
#      vs = [ self.exist() for vi in xrange(low, high+1) ]
      vs = [ self.exist() for vi in xrange(dom) ]
      ovars = atmost_one(self, vs, self.ttt())
      return FDVar(vs, [ ~v for v in ovars ])

##  def add(self, f):
##    self.formulae.append(f)

  def addClause(self, c):
    self.clauses.append(c)

  def to_qdimacs(self):
    ## Anything not mentioned in the prefix is existentially quantified.
    nvars = sum(len(block[1]) for block in self.blocks)
    print "p cnf {0} {1}".format(nvars, len(self.clauses))

    for (qf, bvars) in self.blocks:
      print QBF.QF_repr[qf] + " " + " ".join(map(str, bvars)) + " 0"

    for cl in self.clauses:
      cl_repr = " ".join(map(repr, cl)) + " 0"
      print cl_repr

  ## Construct a non-quantified problem
  def to_dimacs(self):
    assert len(self.blocks) == 1
    
    #print >> sys.stderr, "clauses:", self.clauses, "\n"
    #print >> sys.stderr, "blocks:", self.blocks, "\n"
    nvars = sum(len(block[1]) for block in self.blocks)
    #print >> sys.stderr, "sumli nvar len:", nvars, len(self.clauses),"\n"
    
    
    
    print >> sys.stderr, "p cnf {0} {1}".format(nvars, len(self.clauses))
    print "p cnf {0} {1}".format(nvars, len(self.clauses))
    
    
    
    #print "p cnf 33 71"
    for cl in self.clauses:
      cl_repr = " ".join(map(repr, cl)) + " 0"
      #print >> sys.stderr, cl_repr
      print cl_repr

